/*
 * Implementation of the ANTS client library
 *
 * $Id: antslib.c,v 1.9 2004/06/25 01:29:25 jakob Exp $
 *
 *  Copyright (C) 2001 Jakob Oestergaard <jakob@unthought.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#include "antslib.h"
#include "config.h"

#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#if 0
# define MAX_ANTSD_STREAM  65535
# define RETRY_DELAY 500
# define JOBNET_PORT htons(86)
/*#define JOBNET_PORT htons(1086)*/
# define PRINT_INTERVAL 10
#endif

const char * antsd_socket = "/tmp/antsd";
static int revsock = -1;


/*
 * Insist on reading the given amount of data, unless an error occurs
 * of course
 */
static int blocking_read(int sock, void *buf, size_t len)
{
  int olen = len;
  char *bptr = buf;
  while (len) {
    int rc = recv(sock, bptr, len, MSG_NOSIGNAL);
    if (rc == 0)
      return 0;
    if (rc == -1) {
      if (errno == EINTR)
	continue;
      return -1;
    }
    len -= rc;
    bptr += rc;
  }
  return olen;
}

int ants_local_connect(void)
{
  int antsd;
  int rc;
  struct sockaddr_un addr;
  int opt_val;

  /* Connect */
  antsd = socket(AF_UNIX, SOCK_STREAM, 0);
  if(antsd == -1)
    return -1;

  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, antsd_socket);

  opt_val = 1;
  rc = setsockopt(antsd, SOL_SOCKET, SO_PASSCRED, &opt_val, sizeof(opt_val));
  if(rc)
    return -1;

  rc = connect(antsd, (struct sockaddr*)&addr, sizeof(struct sockaddr_un));
  if(rc) 
    return -1;

  return antsd;
}


int ants_imclient_best_host(struct alloc_reply* response,
			    int antsd, const char* type)
{
  int rc;
  struct antsd_request req;
  unsigned to_recv;
  int tport;
  int bound_port = -1;

  /*
   * Die if type is too long
   */
  if(strlen(type) >= TYPE_LEN) {
    return 1;
  }

  /*
   * Cleanup
   */
  if(revsock != -1) {
    close(revsock);
  }

  /*
   * Bind to some port where we can listen for the antsdph
   * reverse connection
   */
  revsock = socket(AF_INET, SOCK_STREAM, 0);
  if(revsock == -1) {
    // Shoot !  We're definitely out of resources now!
    return 2;
  }

  while(bound_port == -1) {
    int brc;
    struct sockaddr_in subaddr;
    memset(&subaddr, 0, sizeof(subaddr));
    subaddr.sin_family = AF_INET;
    subaddr.sin_addr.s_addr = INADDR_ANY;
    tport = rand() % (REVERSE_PORT_END - REVERSE_PORT_BEGIN);
    tport += REVERSE_PORT_BEGIN;
    subaddr.sin_port = htons(tport);

    brc = bind(revsock, (const struct sockaddr*)&subaddr,
	       sizeof(struct sockaddr_in));

    if(brc == 0)
      bound_port = tport;
  }
  listen(revsock, 3);

  /*
   * Tell ANTS that we're an immediate client and we
   * want to know about best hosts
   */
  req.request = 0;
  strcpy(req.data.joballoc.jobtype, type);
  req.data.joballoc.revport = htons(tport);

  rc = send(antsd, &req, sizeof(req), MSG_NOSIGNAL);
  if(rc != sizeof(req)) 
    return -1;

  /*
   * Await response
   */
  to_recv = sizeof(*response);
  rc = blocking_read(antsd, response, to_recv);
  if(rc != to_recv)
    return -1;

  return 0;
}


int ants_imclient_connect(int antsd, const struct alloc_reply* alloc)
{
  /*
   * Close old connection
   */
  close(antsd);

  /*
   * Listen for incoming connections, but only accept connections
   * from alloc->address
   */
  while(1) {
    struct sockaddr_in padr;
    socklen_t adrlen = sizeof(padr);
    int arc = accept(revsock, (struct sockaddr*)&padr, &adrlen);
    if(arc == -1) {
      /*
       * Accept failed - exit with -1 and have caller read errno
       */
      return -1;
    }
    if(adrlen != sizeof(struct sockaddr_in)) {
      /*
       * Length mismatch - return -1 and let client figure out
       * something went wrong although errno == 0
       */
      return -1;
    }
    /*
     * If address matches, accept the connection
     */
    if(padr.sin_addr.s_addr == alloc->address) {
      antsd = arc;
      break;
    }
    /*
     * Otherwise, ditch the connection and keep listening
     */
    close(arc);
    listen(revsock, 3);
  }

  /*
   * Ok, we're done - now stop listening!
   */
  close(revsock);
  revsock = -1;

  return antsd;
}

int ants_imclient_submit(int antsd, const char * job_type,
			 const struct alloc_reply * alloc,
			 const char * work_dir,
			 const char * command_string,
			 unsigned cs_length,
			 int transferenv)
{
  struct alloc_auth auth;
  int src, rc;
  /*
   * Send: [type] [seq] [working directory]
   */
  strncpy(auth.type, job_type, TYPE_LEN);
  auth.serial = alloc->serial;

  strncpy(auth.workdir, work_dir, CWD_LENGTH);

  src = send(antsd, &auth, sizeof(auth), MSG_NOSIGNAL);
  if(src != sizeof(auth))
    return -1;

  /*
   * If our serial number wasn't accepted by the server, or if the job
   * type is not available on the server, or if the server for some
   * other reason decides that it doesn't like us, it will simply
   * close the connection now.
   *
   * Meaning; the next send() may fail. Or we may first discover what
   * happened later on.
   */

  /*
   * Send command to antsd
   */ 
  rc = send(antsd, &cs_length, sizeof(cs_length), MSG_NOSIGNAL);
  if(rc < 0) return -1;
  rc = send(antsd, command_string, cs_length, MSG_NOSIGNAL);
  if(rc < 0) return -1;

  /*
   * Should we transfer environment?
   */
  if (!transferenv) {
    /* nope - send 0 env. strings */
    unsigned count = 0;
    rc = send(antsd, &count, sizeof(count), MSG_NOSIGNAL);
    if(rc < 0)
      return -1;
  } else {
    /*
     * yep - go ahead!
     *
     * We traverse thru the environ string list, first to count the
     * variables we want to send, then to actually send them
     */
    unsigned count = 0;
    char **envp = environ;
    while (*envp++) ++count;
    rc = send(antsd, &count, sizeof(count), MSG_NOSIGNAL);
    if(rc < 0) return -1;

    for (envp = environ; *envp; ++envp) {
      /* Send length */
      unsigned elen = strlen(*envp);
      rc = send(antsd, &elen, sizeof(elen), MSG_NOSIGNAL);
      if(rc < 0) return -1;
      /* Send data */
      rc = send(antsd, *envp, elen, MSG_NOSIGNAL);
      if(rc < 0) return -1;
    }
    /* that's it and that's that */
  }

  return 0;
}


ants_message_t ants_imclient_readmsg(struct cleanstring * stringdata,
				     int * intdata, int antsd)
{
  struct procreply reply;
  int rc;

  if ((rc = blocking_read(antsd, &reply, sizeof(reply))) == sizeof(reply)) {
    /*
     * Now, see what command we might have gotten
     */
    switch(reply.cmd) {
    case PROCREPLY_STDOUT:  /* stdout */
      if (!(stringdata->str = malloc(reply.misc)))
	return ANTS_MSG_ERROR;
      stringdata->len = reply.misc;
      memcpy(stringdata->str, reply.buffer, reply.misc);
      return ANTS_MSG_STDOUT;
    case PROCREPLY_STDERR:  /* stderr */
      if (!(stringdata->str = malloc(reply.misc)))
	return ANTS_MSG_ERROR;
      stringdata->len = reply.misc;
      memcpy(stringdata->str, reply.buffer, reply.misc);
      return ANTS_MSG_STDERR;
    case PROCREPLY_EXITED: { /* exit - with status */
      *intdata = reply.misc;
      return ANTS_MSG_EXITVAL;
    }
    case PROCREPLY_SIGDIE: { /* exit - with signal */
      *intdata = reply.misc;
      return ANTS_MSG_EXITSIG;
    }
    default:
      return ANTS_MSG_ERROR;
    }
  }

  return ANTS_MSG_CLOSED;
}

int ants_imclient_sendstdin(int len, char * stringdata, int antsd)
{
  struct rantcommand msg;
  int ofs = 0;

  /*
   * Special-casing for length=0 (stdin closed)
   */
  if (len == 0) {
    msg.cmd = 0;
    msg.misc = 0; /* length */
    memset(msg.buffer, 0, sizeof(msg.buffer));

    int rc = send(antsd, &msg, sizeof(msg), 0);
    if (rc != sizeof(msg))
      return -1;
  }

  /*
   * The usual case: we have real data to send
   */
  while (len > 0) {
    unsigned tocopy;
    int rc;

    /* setup command - 0: stdin */
    msg.cmd = 0;

    /* setup message */
    tocopy = len < MAX_ANTSD_MESSAGE ? len : MAX_ANTSD_MESSAGE;
    memcpy(msg.buffer, stringdata + ofs, tocopy);
    len -= tocopy;
    ofs += tocopy;

    /* length */
    msg.misc = tocopy;

    /* send */
    rc = send(antsd, &msg, sizeof(msg), 0);

    if (rc != sizeof(msg))
      return -1;
  }

  return 0;
}


int ants_imclient_send_signal(int signal, int antsd)
{
  int rc;
  struct rantcommand msg;
  msg.cmd = 1; // Send signal
  msg.misc = signal;
  memset(msg.buffer, 0, sizeof(msg.buffer));

  rc = send(antsd, &msg, sizeof(msg), 0);
  if (rc != sizeof(msg))
    return -1;

  return 0;
}


int ants_statclient(int antsd)
{
  int rc;
  struct antsd_request req;

  /*
   * Tell ANTS that we're status requesting client and we want to know
   * about status updates
   */
  req.request = 1;
  rc = send(antsd, &req, sizeof(req), MSG_NOSIGNAL);
  if(rc != sizeof(req)) 
    return -1;

  return 0;
}

