/*
 * ANTS Client Library
 *
 * $Id: antslib.h,v 1.9 2004/06/22 15:26:59 jakob Exp $
 *
 *  Copyright (C) 2001,2004 Jakob Oestergaard <jakob@unthought.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#ifndef __ANTSLIB_H
#define __ANTSLIB_H

#ifdef __cplusplus
extern "C" {
#endif

# define CWD_LENGTH 1024
# define TYPE_LEN 16
# define LOG_LEN 1024
# define MAX_ANTSD_MESSAGE 1024

  typedef enum { ANTS_MSG_ERROR = 0,  /* bad output from ANTS */
		 ANTS_MSG_CLOSED,     /* remote connection closed */
		 ANTS_MSG_STDOUT,     /* standard output      */
		 ANTS_MSG_STDERR,     /* std error output     */
		 ANTS_MSG_EXITVAL,    /* exit with error code */
		 ANTS_MSG_EXITSIG     /* exit with signal     */
  } ants_message_t;

  struct antsd_request {
    /* request:
     *  set to 0 for joballoc,
     *  set to 1 for status gathering
     *  set to 2 for queue requests
     */
    unsigned request;
    union {
      /* Not used for status gathering */
      struct {
	char     jobtype[TYPE_LEN];
	unsigned short revport;
      } joballoc;
    } data;
  };

  struct alloc_reply {
    /*
     * Client can retry no matter which of
     * the errors it receives - all can be
     * temporary.  However, NO_JOB_TYPE is
     * most likely permanent...
     */
# define STATUS_NO_BEST_HOST 0
# define STATUS_NO_JOB_TYPE  1
# define STATUS_ALLOC_ERROR  2
# define STATUS_BEST_HOST    10
# define STATUS_BEST_LOCAL   11
    unsigned status;
    unsigned serial;
    unsigned address;
  };

  struct alloc_auth {
    char type[TYPE_LEN];
    unsigned serial;
    char workdir[CWD_LENGTH];
  };

  struct cleanstring {
    int len;    /* length of string */
    char *str;  /* antslib allocated string - free with free() after use */
  };

  /*
   * Message sent from RANT to ANTSD sub-process, used for
   * signalling the sub-proc, or sending stdin data
   */
  struct rantcommand {
    unsigned cmd;
    unsigned misc;
    char buffer[MAX_ANTSD_MESSAGE];
  };

  /*
   * Message sent from ANTSD sub-process to RANT, used for
   * transmitting back stdout, stderr and exit-code etc.
   */
  struct procreply {
# define PROCREPLY_STDOUT 0
# define PROCREPLY_STDERR 1
# define PROCREPLY_EXITED 2
# define PROCREPLY_SIGDIE 3
    unsigned cmd;
    unsigned misc; /* strlen for stdout/stderr, exitcode or signal otherwise */
    char buffer[MAX_ANTSD_MESSAGE];
  };

  /*
   * Connect to local ANTS daemon, and return
   * socket descriptor.  Returns -1 on error
   * and sets errno accordingly.
   */
  int ants_local_connect();

  /*
   * For ANTS_IMCLIENT:
   *  Request a best-host for job-type. Returns -1
   *  on error and sets errno accordingly. Returns 1
   *  on argument errors (eg. too long job_type etc.)
   *
   *  If result is zero the alloc_reply can hold a status
   *  with the following special meanings:
   *   0 : currently no best host for job type
   *  -1 : we don't know that job-type
   *   1 : invalid job-type argument
   *   2 : unable to create listening socket
   */
  int ants_imclient_best_host(struct alloc_reply* alloc, int antsd,
			      const char* job_type);

  /*
   * Connect to ANTS daemon given in alloc_reply, for
   * immediate job submission. Returns -1 on error
   * and sets errno accordingly.  Returns the new
   * socket descriptor otherwise.
   */
  int ants_imclient_connect(int antsd, const struct alloc_reply* alloc);

  /*
   * Authenticate and submit job to ANTS daemon
   * Returns -1 on syscall error, with errno set appropriately
   * Returns 1 on authentication error
   * Returns 0 on success
   *
   * command_string has the format:
   *   0012[arg_0]\0[arg_1]\0....[arg_11]\0
   * four hex digits telling how many arguments will follow,
   * then the arguments separated with \0 characters.
   */
  int ants_imclient_submit(int antsd, const char * job_type,
			   const struct alloc_reply * alloc,
			   const char * work_dir,
			   const char * command_string,
			   unsigned cs_length,
			   int transferenv);

  /*
   * Process messages from ANTS daemon
   * Returns the message type, and
   * fills in either the stringdata or intdata, depending on message type.
   */
  ants_message_t ants_imclient_readmsg(struct cleanstring *stringdata,
				       int * intdata, int antsd);

  /*
   * Send standard input to the remotely executing process
   *
   * Returns 0 on apparent success, -1 on immediate error.
   */
  int ants_imclient_sendstdin(int len, char * stringdata, int antsd);

  /*
   * Send a signal to running job
   * Returns 0 on success, -1 on error
   */
  int ants_imclient_send_signal(int signal, int antsd);
  
  /*
   * Tell ANTS that we're a status monitoring client
   * and we want job status changes fed back
   */
  int ants_statclient(int ants);


#ifdef __cplusplus
}
#endif

#endif

