///////////////////////////////////////////////////////////////////////////
//
// ANTS meter
//
// text-mode program to show the number of jobs running
//

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "antslib.h"

// XXX: How to detect if theese are present?
#include <sys/socket.h>
#include <curses.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <map>
#include <set>
#include <string>
#include <stdlib.h>
#include <iostream>

// Connection to local antsd
int usock = -1;

bool should_exit = false;

// Information about one job type at one node
struct job_t {
  unsigned max, preferred;
  unsigned running, reserved;
};

// Information about one cluster node
struct node_t {
  unsigned id;
  std::string name;
  std::map<std::string, job_t> jobs;
};

// Information about all known nodes
std::map<unsigned, node_t> nodes;

// The set of all known jobtypes
std::set<std::string> jobtypes;

bool init()
{
  /*
   * Make sure we're connected to local antsd
   */
  if(usock == -1) {
    int rc;

    usock = ants_local_connect();
    if(usock < 0) {
      usock = -1;
      /* bad ! */
      std::cerr << "Yuck! connect() failed" << std::endl;
      return false;
    }

    rc = ants_statclient(usock);
    if(rc != 0) {
      std::cerr <<"Couldn't tell ANTS that we're a statclient" << std::endl;
      close(usock);
      usock = -1;
      return false;
    }
  }

  // Curses stuff
  initscr(); cbreak(); noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
  // Clear screen for a start
  move(0, 0);
  erase();
  refresh();

  return true;
}

void done()
{
  // Curses stuff
  endwin();
  
  if(usock == -1) return;
  
  close(usock);
}

void update()
{
  static char * buf = 0;
  int rc;

  unsigned peer_id;
  char jobtype[TYPE_LEN];
  job_t job;

  fd_set readfds, writefds;
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_SET(usock, &readfds);
  int nreads;
  nreads = select(usock + 1, &readfds, &writefds, 0, 0);


  /*
   * Check buffer alloc.
   */
  if(!buf) {
    buf = (char*)malloc(LOG_LEN);
    if(!buf) return;
  }

  /*
   * Read and parse
   */
  rc = recv(usock, buf, LOG_LEN, 0);
  if(rc <= 0) {
    printw("Yuck! recv() failed", buf);
    //kill_local_socket();
    return;
  }

  rc = sscanf(buf, "%u %15s %u %u %u %u",
	      &peer_id, jobtype, 
	      &job.max, &job.preferred,
	      &job.running, &job.reserved);
  if(rc != 6) {
    // This can happen, as there is no guaranty about preservation of
    // record boundaries on unix sockets.

    //printw("Yuck! sscanf() failed: [%s]\n", buf);
    return;
  }

  nodes[peer_id].jobs[jobtype] = job;
  jobtypes.insert(jobtype);
}

void print()
{
  move(0, 0);
  erase();
  for (std::set<std::string>::const_iterator it = jobtypes.begin();
       it != jobtypes.end(); ++it) {
    printw("\t%s", it->c_str());
  }
  printw("\n");
  for (std::map<unsigned, node_t>::const_iterator node = nodes.begin();
       node != nodes.end(); ++node) {
    printw("Node:");
    for (std::set<std::string>::const_iterator it = jobtypes.begin();
	 it != jobtypes.end(); ++it) {
      std::map<std::string, job_t>::const_iterator job
	= node->second.jobs.find(*it);
      if (job != node->second.jobs.end()) {
#if 0
	printw("\t%u", job->second.running);
#else
	printw("\t");
	unsigned i = 0;
	for (; i < job->second.running && i < 7; i++)
	  if (i < job->second.preferred)
	    printw("*");
	  else
	    printw("!");
	for (; i < job->second.running + job->second.reserved && i < 7; i++)
	  printw("r");
	for (; i < job->second.preferred && i < 7; i++)
	  printw(".");
	for (; i < job->second.max && i < 7; i++)
	  printw(".");
      } else
	printw("\t");
#endif
    }
    printw("\n");
  }
  refresh();
}

int main()
{
  if (init()) {
    while (!should_exit) {
      update();
      print();
    }
    done();
  }
}
