/*
 * The Remote ANTs execution utility
 *
 * $Id: rant.cc,v 1.15 2004/10/20 11:04:13 jrk Exp $
 *
 *  Copyright (C) 2000, 2001, 2004 Jakob Oestergaard <jakob@unthought.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#include "config.h"
#include "antslib.h"

#include <iostream>
#include <string.h>
#include <cstdlib>
#include <list>
#include <cerrno>
#include <sys/time.h>

#include <unistd.h>
#include <signal.h>

// #ifndef HAVE_SSTREAM
// # error Cannot yet build without sstream
// #else
# include <sstream>
# include <iomanip>
// #endif

//
// We need a slick little timer
//
struct Timer {
  void tic();
  void toc();
  double get_delay() const;
  struct timeval data;
};

void usage();
void version();
void fail(const char*);
void errno_fail(const char*);
void signal_handler(int i);
void treat_remote();
void treat_stdin();

int antsd = -1;
int signal_to_send = -1;
int closed_stdin = 0;

int main(int argc, char** argv)
{
  std::string job_type;
  std::string command;
  std::list<std::string> arg_list;

  //
  // First of all, parse the command line.
  // It can hold:
  //     rant [-t <type>] <command> <args...>
  // or: rant -v
  //
  if(argc <= 1)
    usage();

  if(geteuid() < PROCESS_MIN_UID)
    fail("User ID is too low - ANTS will refuse to service our requests");

  int carg_ndx = 1;
  while(carg_ndx != argc) {

    std::string carg = argv[carg_ndx];

    if(command.size() == 0 && carg == "-v") {
      version();
    }

    if(command.size() == 0 && carg == "-t") {
      // Now, next arg must hold a job type
      carg_ndx ++;
      if(carg_ndx == argc)
	fail("Nothing supplied to -t argument");

      job_type = argv[carg_ndx];

      carg_ndx ++;
      continue;
    }

    // Ok, this must be either a command or an argument
    if(command.size() == 0) {
      command = carg;
      
      if(job_type.size() == 0)
	job_type = command;

      carg_ndx ++;
      continue;
    }

    // It's an argument then - since we have the command already
    arg_list.push_back(carg);
    carg_ndx ++;
    continue;
  }


  //
  // Setup connection with local ANTS daemon
  //
  antsd = ants_local_connect();

  if(antsd == -1)
    errno_fail("Unable to connect to local ANTS daemon");

  //
  // Request best host for job-type
  //
  // If there is no best host for given job-type, wait a little, then re-try
  //
  struct alloc_reply alloc_response;

  // Have we written "Temporary job allocation failure..."
  bool temp_failure = false;

  do {
  //
  // Start timing this
  //
  Timer t_start;
  t_start.tic();

#if 0
  struct timeval tv;
  gettimeofday(&tv, 0);
  std::cerr << "jobslot begin at : " << tv.tv_sec 
	    << ":" << tv.tv_usec << std::endl;
#endif

  int alloc_rc = ants_imclient_best_host(&alloc_response, antsd,
					 job_type.c_str());

  //
  // End timing
  //
#if 0
  gettimeofday(&tv, 0);
  std::cerr << "jobslot finish at : " << tv.tv_sec
	    << ":" << tv.tv_usec << std::endl;

  t_start.toc();
  std::cerr << "Scheduling latency: " << t_start.get_delay() * 1000
	    << " msec." << std::endl;
#endif


    if(alloc_rc < 0)
      errno_fail("Unable to query local ANTS daemon for best host");
    if(alloc_rc > 0)
      fail("Internal error when querying local ANTS daemon for best host");

    //
    // Handle various errors
    //
    switch(alloc_response.status) {
    case STATUS_NO_BEST_HOST:
      std::cerr << "Currently no best host for \"" << job_type
		<< "\" job. Will retry..." << std::endl;
      temp_failure = true;
      sleep(10);
      continue;
    case STATUS_NO_JOB_TYPE:
      std::cerr << "Job-type \"" << job_type << "\" is unknown" << std::endl;
      return 1;
    case STATUS_ALLOC_ERROR:
      std::cerr << "Temporary job allocation failure. Will retry..." << std::endl;
      temp_failure = true;
      sleep(10);
      continue;
    }

    // Exit loop if best host found
    if(alloc_response.status == STATUS_BEST_HOST
       || alloc_response.status == STATUS_BEST_LOCAL)
      break;

  } while(true);

  if (temp_failure)
    std::cerr << "Job allocation success." << std::endl;

  //
  // Now, drop the old antsd connection and connect to the
  // best host for the job
  //
  antsd = ants_imclient_connect(antsd, &alloc_response);
  if(antsd == -1)
    errno_fail("Unable to connect to ANTS daemon on best-host");


  //
  // Get CWD for the command
  //
  char * cwd = new char[CWD_LENGTH];
  if(!getcwd(cwd, CWD_LENGTH))
    fail("Couldn't get current working directory");

  //
  // Now we should build a string holding the entire command
  // line we want to execute
  //
  // That string is encoded like:
  //   0012[arg_0]\0[arg_1]\0....[arg_11]\0
  // four hex digits telling how many
  // arguments will follow, then the arguments
  // separated with \0 characters.
  //
  
  std::ostringstream cmd_buffer;

  // Print element count as four hex digits
  cmd_buffer << std::hex << std::setw(4) << std::setfill('0')
	     << 1 + arg_list.size();

  // Add command
  cmd_buffer << command << '\0';

  // Add arguments
  for(std::list<std::string>::const_iterator ali = arg_list.begin();
      ali != arg_list.end(); ali++)
    cmd_buffer << *ali << '\0';

  int submit_rc = ants_imclient_submit(antsd, job_type.c_str(),
				       &alloc_response,
				       cwd,
				       cmd_buffer.str().c_str(),
				       cmd_buffer.str().size(), 1);

  if(submit_rc < 0)
    errno_fail("Job submission failed");
  if(submit_rc > 0)
    fail("Job submission failed - possibly authentication error");

  //
  // Setup signal handlers, signals will be forwarded to the remote process
  //
  struct sigaction act;
  memset(&act, 0, sizeof(act));
  act.sa_handler = signal_handler;
  if (sigaction(SIGHUP, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGINT, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGQUIT, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGTERM, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGUSR1, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGUSR2, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGTSTP, &act, 0)) fail("Cannot set signal handler");
  if (sigaction(SIGCONT, &act, 0)) fail("Cannot set signal handler");

  //
  // Now, process ANTS messages
  //
  do {

    //
    // We want to wait for either stdin, a signal, or a remote message
    //
    fd_set input;
    FD_ZERO(&input);
    fd_set output;
    FD_ZERO(&output);
    fd_set except;
    FD_ZERO(&except);

    if (!closed_stdin)
      FD_SET(0, &input);    // wait for stdin read

    FD_SET(antsd, &input);  // wait for remote antsdph read
    FD_SET(antsd, &except); // wait for remote error

    int selrc = select(antsd + 1, &input, &output, &except, 0);

    // Did select fail?
    if (selrc == -1 && errno != EINTR) {
      std::cerr << "Error listening for stuff to do - weird.." << std::endl;
      exit(1);
    }

    // Ok, should we read from remote?
    if (selrc > 0 && (FD_ISSET(antsd, &input) || FD_ISSET(antsd, &except)))
      treat_remote();

    // Should we read from stdin?
    if (!closed_stdin && selrc > 0 && FD_ISSET(0, &input))
      treat_stdin();

    // Should we send a signal?
    if (signal_to_send != -1) {
      ants_imclient_send_signal(signal_to_send, antsd);
      signal_to_send = -1;
    }

  } while(true);

  return 0;
}

void usage(void)
{
  std::cerr << "Usage:" << std::endl
	    << " spawn job:  rant [-t <type>] <command> <args...>" << std::endl
	    << std::endl;
  exit(1);
}

void version()
{
  std::cerr << "ANTS (rant) v. " << ANTS_VERSION << std::endl
	    << "(C) 2000, 2001, 2002 by J. Oestergaard, " << std::endl
	    << " http://unthought.net, jakob@unthought.net" << std::endl
	    << "-------------------------------------------" << std::endl
	    << "Licensed under the GNU Public License." << std::endl
	    << std::endl;
  exit(0);
}

void signal_handler(int signal)
{
  // We have received a signal, send it to the remote process
  signal_to_send = signal;
}

void treat_remote()
{
  cleanstring string_buffer;
  int int_buffer;
  ants_message_t msg = ants_imclient_readmsg(&string_buffer,
					     &int_buffer, antsd);
  switch(msg) {
  case ANTS_MSG_STDOUT:
    std::cout << std::string(string_buffer.str, string_buffer.len)
	      << std::flush;
    free(string_buffer.str);
    break;
  case ANTS_MSG_STDERR:
    std::cerr <<  std::string(string_buffer.str, string_buffer.len)
	      << std::flush;
    free(string_buffer.str);
    break;
  case ANTS_MSG_EXITVAL:
    exit(int_buffer);
  case ANTS_MSG_EXITSIG:
    std::cerr << "rant warning: remote job terminated " << std::endl
	      << "              with signal " << int_buffer << std::endl
	      << "              I will too." << std::endl;
    // Reset handler for this signal to the default one, before
    // raising the signal
    struct sigaction act;
    memset(&act, 0, sizeof(act));
    act.sa_handler = SIG_DFL;
    sigaction(int_buffer, &act, 0); // Do not care about errors
    raise(int_buffer);
    exit(1);			// We should not get here
  case ANTS_MSG_CLOSED:
    std::cerr << "Remote job terminated, without signal or error code"
	      << std::endl;
    exit(1);
  default:
    std::cerr << "Unknown return value (" << int(msg)
	      << ") from remote ANTS" << std::endl;
    exit(1);
  }
}

void treat_stdin()
{
  // Just try to read one 'message size' of dat
  char buffer[MAX_ANTSD_MESSAGE];
  int rc = read(0, buffer, MAX_ANTSD_MESSAGE);
  if (rc == -1)
    return;
  if (rc == 0) {
    //
    // Read returned 0 on stdin - close stdin and tell remote end to
    // do the same
    //
    ants_imclient_sendstdin(0, 0, antsd);
    // Remember this for our own select()
    closed_stdin = 1;
    close(0);
    return;
  }
  if (ants_imclient_sendstdin(rc, buffer, antsd))
    std::cerr << "Error sending stdin..." << std::endl;
}


void fail(const char* e) 
{
  std::cerr << "Error:" << std::endl
	    << " " << e << std::endl
	    << std::endl;
  exit(2);
}

void errno_fail(const char* e) 
{
  std::cerr << "Error:" << std::endl
	    << " " << e << std::endl
	    << "errno = " << strerror(errno) << std::endl
	    << std::endl;
  exit(1);
}

//
// Start timing - store current time in data
//
void Timer::tic()
{
  gettimeofday(&data, 0);
}

//
// Stop timing - store the interval in data
//
void Timer::toc()
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  data.tv_sec = tv.tv_sec - data.tv_sec;
  data.tv_usec = tv.tv_usec - data.tv_usec;
}

//
// Return delay as FP [secs]
//
double Timer::get_delay() const
{
  double ret = data.tv_usec / 1E6;
  ret += data.tv_sec;
  return ret;
}
